from autogen.agentchat import ConversableAgent, Agent
from ..agent import AgentAsync

from .. import system
from .. import constants

import asyncio
from typing import Callable, Optional, List, Dict, Union, Any, Tuple


def create_memgpt_autogen_agent_from_config(
    name: str,
    system_message: Optional[str] = "You are a helpful AI Assistant.",
    is_termination_msg: Optional[Callable[[Dict], bool]] = None,
    max_consecutive_auto_reply: Optional[int] = None,
    human_input_mode: Optional[str] = "TERMINATE",
    function_map: Optional[Dict[str, Callable]] = None,
    code_execution_config: Optional[Union[Dict, bool]] = None,
    llm_config: Optional[Union[Dict, bool]] = None,
    default_auto_reply: Optional[Union[str, Dict, None]] = "",
):
    """
    TODO support AutoGen config workflow in a clean way with constructors 
    """
    raise NotImplementedError


class MemGPTAgent(ConversableAgent):

    def __init__(
        self,
        name: str,
        agent: AgentAsync,
        skip_verify=False
    ):
        super().__init__(name)
        self.agent = agent
        self.skip_verify = skip_verify
        self.register_reply([Agent, None], MemGPTAgent._a_generate_reply_for_user_message)
        self.register_reply([Agent, None], MemGPTAgent._generate_reply_for_user_message)

    def _generate_reply_for_user_message(
        self,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Agent] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None]]:
        return asyncio.run(self._a_generate_reply_for_user_message(messages=messages, sender=sender, config=config))

    async def _a_generate_reply_for_user_message(
        self,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Agent] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None]]:
        ret = []
        # for the interface
        self.agent.interface.reset_message_list()

        for msg in messages:
            user_message = system.package_user_message(msg['content'])
            while True:
                new_messages, heartbeat_request, function_failed, token_warning = await self.agent.step(user_message, first_message=False, skip_verify=self.skip_verify)
                ret.extend(new_messages)
                # Skip user inputs if there's a memory warning, function execution failed, or the agent asked for control
                if token_warning:
                    user_message = system.get_token_limit_warning()
                elif function_failed:
                    user_message = system.get_heartbeat(constants.FUNC_FAILED_HEARTBEAT_MESSAGE)
                elif heartbeat_request:
                    user_message = system.get_heartbeat(constants.REQ_HEARTBEAT_MESSAGE)
                else:
                    break

        # Pass back to AutoGen the pretty-printed calls MemGPT made to the interface
        pretty_ret = MemGPTAgent.pretty_concat(self.agent.interface.message_list)
        return True, pretty_ret

    @staticmethod
    def pretty_concat(messages):
        """AutoGen expects a single response, but MemGPT may take many steps.
        
        To accommodate AutoGen, concatenate all of MemGPT's steps into one and return as a single message.
        """
        ret = {
            'role': 'assistant',
            'content': ''
        }
        lines = []
        for m in messages:
            lines.append(f"{m}")
        ret['content'] = '\n'.join(lines)
        return ret



import os


def get_system_text(key):
    filename = f'{key}.txt'
    file_path = os.path.join(os.path.dirname(__file__), 'system', filename)

    if os.path.exists(file_path):
        with open(file_path, 'r') as file:
            return file.read().strip()
    else:
        raise FileNotFoundError(f"No file found for key {key}, path={file_path}")



import faiss
from glob import glob
from tqdm import tqdm
import numpy as np
import argparse
import json

def build_index(embedding_files: str,
                index_name: str):

    index = faiss.IndexFlatL2(1536)
    file_list = sorted(glob(embedding_files))

    for embedding_file in file_list:
        print(embedding_file)
        with open(embedding_file, 'rt', encoding='utf-8') as file:
            embeddings = []
            l = 0
            for line in tqdm(file):
                # Parse each JSON line
                data = json.loads(line)
                embeddings.append(data)
                l += 1
            data = np.array(embeddings).astype('float32')
            print(data.shape)
            try:
                index.add(data)
            except Exception as e:
                print(data)
                raise e

    faiss.write_index(index, index_name)


if __name__ ==  '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument('--embedding_files', type=str, help='embedding_filepaths glob expression')
    parser.add_argument('--output_index_file', type=str, help='output filepath')
    args = parser.parse_args()    

    build_index(
        embedding_files=args.embedding_files,
        index_name=args.output_index_file
    )


import json

from .utils import get_local_time
from .constants import INITIAL_BOOT_MESSAGE, INITIAL_BOOT_MESSAGE_SEND_MESSAGE_THOUGHT, INITIAL_BOOT_MESSAGE_SEND_MESSAGE_FIRST_MSG, MESSAGE_SUMMARY_WARNING_STR


def get_initial_boot_messages(version='startup'):

    if version == 'startup':
        initial_boot_message = INITIAL_BOOT_MESSAGE
        messages = [
            {"role": "assistant", "content": initial_boot_message},
        ]

    elif version == 'startup_with_send_message':
        messages = [
            # first message includes both inner monologue and function call to send_message
            {
                "role": "assistant",
                "content": INITIAL_BOOT_MESSAGE_SEND_MESSAGE_THOUGHT,
                "function_call": {
                    "name": "send_message",
                    "arguments": "{\n  \"message\": \"" + f"{INITIAL_BOOT_MESSAGE_SEND_MESSAGE_FIRST_MSG}" + "\"\n}"
                }
            },
            # obligatory function return message
            {
                "role": "function",
                "name": "send_message",
                "content": package_function_response(True, None)
            }
        ]

    elif version == 'startup_with_send_message_gpt35':
        messages = [
            # first message includes both inner monologue and function call to send_message
            {
                "role": "assistant",
                "content": "*inner thoughts* Still waiting on the user. Sending a message with function.",
                "function_call": {
                    "name": "send_message",
                    "arguments": "{\n  \"message\": \"" + f"Hi, is anyone there?" + "\"\n}"
                }
            },
            # obligatory function return message
            {
                "role": "function",
                "name": "send_message",
                "content": package_function_response(True, None)
            }
        ]

    else:
        raise ValueError(version)

    return messages


def get_heartbeat(reason='Automated timer', include_location=False, location_name='San Francisco, CA, USA'):

    # Package the message with time and location
    formatted_time = get_local_time()
    packaged_message = {
        "type": 'heartbeat',
        "reason": reason,
        "time": formatted_time,
    }

    if include_location:
        packaged_message["location"] = location_name

    return json.dumps(packaged_message)


def get_login_event(last_login='Never (first login)', include_location=False, location_name='San Francisco, CA, USA'):

    # Package the message with time and location
    formatted_time = get_local_time()
    packaged_message = {
        "type": 'login',
        "last_login": last_login,
        "time": formatted_time,
    }

    if include_location:
        packaged_message["location"] = location_name

    return json.dumps(packaged_message)


def package_user_message(user_message, time=None, include_location=False, location_name='San Francisco, CA, USA'):

    # Package the message with time and location
    formatted_time = time if time else get_local_time()
    packaged_message = {
        "type": 'user_message',
        "message": user_message,
        "time": formatted_time,
    }

    if include_location:
        packaged_message["location"] = location_name

    return json.dumps(packaged_message)

def package_function_response(was_success, response_string, timestamp=None):

    formatted_time = get_local_time() if timestamp is None else timestamp
    packaged_message = {
        "status": 'OK' if was_success else 'Failed',
        "message": response_string,
        "time": formatted_time,
    }

    return json.dumps(packaged_message)


def package_summarize_message(summary, summary_length, hidden_message_count, total_message_count, timestamp=None):

    context_message = \
        f"Note: prior messages ({hidden_message_count} of {total_message_count} total messages) have been hidden from view due to conversation memory constraints.\n" \
        + f"The following is a summary of the previous {summary_length} messages:\n {summary}"

    formatted_time = get_local_time() if timestamp is None else timestamp
    packaged_message = {
        "type": 'system_alert',
        "message": context_message,
        "time": formatted_time,
    }

    return json.dumps(packaged_message)


def package_summarize_message_no_summary(hidden_message_count, timestamp=None, message=None):
    """Add useful metadata to the summary message"""

    # Package the message with time and location
    formatted_time = get_local_time() if timestamp is None else timestamp
    context_message = message if message else \
        f"Note: {hidden_message_count} prior messages with the user have been hidden from view due to conversation memory constraints. Older messages are stored in Recall Memory and can be viewed using functions."
    packaged_message = {
        "type": 'system_alert',
        "message": context_message,
        "time": formatted_time,
    }

    return json.dumps(packaged_message)


def get_token_limit_warning():

    formatted_time = get_local_time()
    packaged_message = {
        "type": 'system_alert',
        "message": MESSAGE_SUMMARY_WARNING_STR,
        "time": formatted_time,
    }

    return json.dumps(packaged_message)


import json
import re

from colorama import Fore, Style, init
init(autoreset=True)


# DEBUG = True  # puts full message outputs in the terminal
DEBUG = False  # only dumps important messages in the terminal


class DummyInterface(object):

    def set_message_list(self, message_list):
        pass

    async def internal_monologue(self, msg):
        pass

    async def assistant_message(self, msg):
        pass

    async def memory_message(self, msg):
        pass

    async def system_message(self, msg):
        pass

    async def user_message(self, msg, raw=False):
        pass

    async def function_message(self, msg):
        pass


class AutoGenInterface(object):
    """AutoGen expects a single action return in its step loop, but MemGPT may take many actions.

    To support AutoGen, we keep a buffer of all the steps that were taken using the interface abstraction,
    then we concatenate it all and package back as a single 'assistant' ChatCompletion response.

    The buffer needs to be wiped before each call to memgpt.agent.step()
    """

    def __init__(self, message_list=None, show_user_message=False, fancy=True):
        self.message_list = message_list
        self.show_user_message = show_user_message
        self.fancy = fancy  # set to false to disable colored outputs + emoji prefixes

    def reset_message_list(self):
        """Clears the buffer. Call before every agent.step() when using MemGPT+AutoGen"""
        self.message_list = []

    async def internal_monologue(self, msg):
        # ANSI escape code for italic is '\x1B[3m'
        message = f'\x1B[3m{Fore.LIGHTBLACK_EX}💭 {msg}{Style.RESET_ALL}' if self.fancy else f'[inner thoughts] {msg}'
        self.message_list.append(message)

    async def assistant_message(self, msg):
        message = f'{Fore.YELLOW}{Style.BRIGHT}🤖 {Fore.YELLOW}{msg}{Style.RESET_ALL}' if self.fancy else msg
        self.message_list.append(message)

    async def memory_message(self, msg):
        message = f'{Fore.LIGHTMAGENTA_EX}{Style.BRIGHT}🧠 {Fore.LIGHTMAGENTA_EX}{msg}{Style.RESET_ALL}' if self.fancy else f'[memory] {msg}'
        self.message_list.append(message)

    async def system_message(self, msg):
        message = f'{Fore.MAGENTA}{Style.BRIGHT}🖥️ [system] {Fore.MAGENTA}{msg}{Style.RESET_ALL}' if self.fancy else f'[system] {msg}'
        self.message_list.append(message)

    async def user_message(self, msg, raw=False):
        if not self.show_user_message:
            return

        if isinstance(msg, str):
            if raw:
                message = f'{Fore.GREEN}{Style.BRIGHT}🧑 {Fore.GREEN}{msg}{Style.RESET_ALL}' if self.fancy else f'[user] {msg}'
                self.message_list.append(message)
                return
            else:
                try:
                    msg_json = json.loads(msg)
                except:
                    print(f"Warning: failed to parse user message into json")
                    message = f'{Fore.GREEN}{Style.BRIGHT}🧑 {Fore.GREEN}{msg}{Style.RESET_ALL}' if self.fancy else f'[user] {msg}'
                    self.message_list.append(message)
                    return

        if msg_json['type'] == 'user_message':
            msg_json.pop('type')
            message = f'{Fore.GREEN}{Style.BRIGHT}🧑 {Fore.GREEN}{msg_json}{Style.RESET_ALL}' if self.fancy else f'[user] {msg}'
        elif msg_json['type'] == 'heartbeat':
            if True or DEBUG:
                msg_json.pop('type')
                message = f'{Fore.GREEN}{Style.BRIGHT}💓 {Fore.GREEN}{msg_json}{Style.RESET_ALL}' if self.fancy else f'[system heartbeat] {msg}'
        elif msg_json['type'] == 'system_message':
            msg_json.pop('type')
            message = f'{Fore.GREEN}{Style.BRIGHT}🖥️ {Fore.GREEN}{msg_json}{Style.RESET_ALL}' if self.fancy else f'[system] {msg}'
        else:
            message = f'{Fore.GREEN}{Style.BRIGHT}🧑 {Fore.GREEN}{msg_json}{Style.RESET_ALL}' if self.fancy else f'[user] {msg}'

        self.message_list.append(message)

    async def function_message(self, msg):

        if isinstance(msg, dict):
            message = f'{Fore.RED}{Style.BRIGHT}⚡ [function] {Fore.RED}{msg}{Style.RESET_ALL}'
            self.message_list.append(message)
            return

        if msg.startswith('Success: '):
            message = f'{Fore.RED}{Style.BRIGHT}⚡🟢 [function] {Fore.RED}{msg}{Style.RESET_ALL}' if self.fancy else f'[function - OK] {msg}'
        elif msg.startswith('Error: '):
            message = f'{Fore.RED}{Style.BRIGHT}⚡🔴 [function] {Fore.RED}{msg}{Style.RESET_ALL}' if self.fancy else f'[function - error] {msg}'
        elif msg.startswith('Running '):
            if DEBUG:
                message = f'{Fore.RED}{Style.BRIGHT}⚡ [function] {Fore.RED}{msg}{Style.RESET_ALL}' if self.fancy else f'[function] {msg}'
            else:
                if 'memory' in msg:
                    match = re.search(r'Running (\w+)\((.*)\)', msg)
                    if match:
                        function_name = match.group(1)
                        function_args = match.group(2)
                        message = f'{Fore.RED}{Style.BRIGHT}⚡🧠 [function] {Fore.RED}updating memory with {function_name}{Style.RESET_ALL}:' \
                            if self.fancy else f'[function] updating memory with {function_name}'
                        try:
                            msg_dict = eval(function_args)
                            if function_name == 'archival_memory_search':
                                message = f'{Fore.RED}\tquery: {msg_dict["query"]}, page: {msg_dict["page"]}' \
                                    if self.fancy else f'[function] query: {msg_dict["query"]}, page: {msg_dict["page"]}'
                            else:
                                message = f'{Fore.RED}{Style.BRIGHT}\t{Fore.RED} {msg_dict["old_content"]}\n\t{Fore.GREEN}→ {msg_dict["new_content"]}' \
                                    if self.fancy else f'[old -> new] {msg_dict["old_content"]} -> {msg_dict["new_content"]}'
                        except Exception as e:
                            print(e)
                            message = msg_dict
                    else:
                        print(f"Warning: did not recognize function message")
                        message = f'{Fore.RED}{Style.BRIGHT}⚡ [function] {Fore.RED}{msg}{Style.RESET_ALL}' \
                            if self.fancy else f'[function] {msg}'
                elif 'send_message' in msg:
                    # ignore in debug mode
                    message = None
                else:
                    message = f'{Fore.RED}{Style.BRIGHT}⚡ [function] {Fore.RED}{msg}{Style.RESET_ALL}' \
                        if self.fancy else f'[function] {msg}'
        else:
            try:
                msg_dict = json.loads(msg)
                if "status" in msg_dict and msg_dict["status"] == "OK":
                    message = f'{Fore.GREEN}{Style.BRIGHT}⚡ [function] {Fore.GREEN}{msg}{Style.RESET_ALL}' \
                        if self.fancy else f'[function] {msg}'
            except Exception:
                print(f"Warning: did not recognize function message {type(msg)} {msg}")
                message = f'{Fore.RED}{Style.BRIGHT}⚡ [function] {Fore.RED}{msg}{Style.RESET_ALL}' \
                    if self.fancy else f'[function] {msg}'

        if message: self.message_list.append(message)



import asyncio
import random
import os
import time

from .local_llm.chat_completion_proxy import get_chat_completion

HOST = os.getenv("OPENAI_API_BASE")
HOST_TYPE = os.getenv("BACKEND_TYPE")  # default None == ChatCompletion

import openai

if HOST is not None:
    openai.api_base = HOST


def retry_with_exponential_backoff(
    func,
    initial_delay: float = 1,
    exponential_base: float = 2,
    jitter: bool = True,
    max_retries: int = 20,
    errors: tuple = (openai.error.RateLimitError,),
):
    """Retry a function with exponential backoff."""

    def wrapper(*args, **kwargs):
        # Initialize variables
        num_retries = 0
        delay = initial_delay

        # Loop until a successful response or max_retries is hit or an exception is raised
        while True:
            try:
                return func(*args, **kwargs)

            # Retry on specified errors
            except errors as e:
                # Increment retries
                num_retries += 1

                # Check if max retries has been reached
                if num_retries > max_retries:
                    raise Exception(
                        f"Maximum number of retries ({max_retries}) exceeded."
                    )

                # Increment the delay
                delay *= exponential_base * (1 + jitter * random.random())

                # Sleep for the delay
                time.sleep(62)

            # Raise exceptions for any errors not specified
            except Exception as e:
                raise e

    return wrapper


@retry_with_exponential_backoff
def completions_with_backoff(**kwargs):
    return openai.ChatCompletion.create(**kwargs)


def aretry_with_exponential_backoff(
    func,
    initial_delay: float = 1,
    exponential_base: float = 2,
    jitter: bool = True,
    max_retries: int = 20,
    errors: tuple = (openai.error.RateLimitError,),
):
    """Retry a function with exponential backoff."""

    async def wrapper(*args, **kwargs):
        # Initialize variables
        num_retries = 0
        delay = initial_delay

        # Loop until a successful response or max_retries is hit or an exception is raised
        while True:
            try:
                return await func(*args, **kwargs)

            # Retry on specified errors
            except errors as e:
                print(f"acreate (backoff): caught error: {e}")
                # Increment retries
                num_retries += 1

                # Check if max retries has been reached
                if num_retries > max_retries:
                    raise Exception(
                        f"Maximum number of retries ({max_retries}) exceeded."
                    )

                # Increment the delay
                delay *= exponential_base * (1 + jitter * random.random())

                # Sleep for the delay
                await asyncio.sleep(62)

            # Raise exceptions for any errors not specified
            except Exception as e:
                raise e

    return wrapper


@aretry_with_exponential_backoff
async def acompletions_with_backoff(**kwargs):
    # Local model
    if HOST_TYPE is not None:
        return await get_chat_completion(**kwargs)

    # OpenAI / Azure model
    else:
        azure_openai_deployment = os.getenv("AZURE_OPENAI_DEPLOYMENT")
        if azure_openai_deployment is not None:
            kwargs["deployment_id"] = azure_openai_deployment
        return await openai.ChatCompletion.acreate(**kwargs)


@aretry_with_exponential_backoff
async def acreate_embedding_with_backoff(**kwargs):
    """Wrapper around Embedding.acreate w/ backoff"""
    azure_openai_deployment = os.getenv("AZURE_OPENAI_DEPLOYMENT")
    if azure_openai_deployment is not None:
        kwargs["deployment_id"] = azure_openai_deployment
    return await openai.Embedding.acreate(**kwargs)


async def async_get_embedding_with_backoff(text, model="text-embedding-ada-002"):
    """To get text embeddings, import/call this function
    It specifies defaults + handles rate-limiting + is async"""
    text = text.replace("\n", " ")
    response = await acreate_embedding_with_backoff(input=[text], model=model)
    embedding = response["data"][0]["embedding"]
    return embedding




from .prompts import gpt_functions
from .prompts import gpt_system
from .agent import AgentAsync
from .utils import printd


DEFAULT = 'memgpt_chat'

def use_preset(preset_name, model, persona, human, interface, persistence_manager):
    """Storing combinations of SYSTEM + FUNCTION prompts"""

    if preset_name == 'memgpt_chat':

        functions = [
            'send_message', 'pause_heartbeats',
            'core_memory_append', 'core_memory_replace',
            'conversation_search', 'conversation_search_date',
            'archival_memory_insert', 'archival_memory_search',
        ]
        available_functions = [v for k,v in gpt_functions.FUNCTIONS_CHAINING.items() if k in functions]
        printd(f"Available functions:\n", [x['name'] for x in available_functions])
        assert len(functions) == len(available_functions)

        if 'gpt-3.5' in model:
            # use a different system message for gpt-3.5
            preset_name = 'memgpt_gpt35_extralong'

        return AgentAsync(
            model=model,
            system=gpt_system.get_system_text(preset_name),
            functions=available_functions,
            interface=interface,
            persistence_manager=persistence_manager,
            persona_notes=persona,
            human_notes=human,
            # gpt-3.5-turbo tends to omit inner monologue, relax this requirement for now
            first_message_verify_mono=True if 'gpt-4' in model else False,
        )

    else:
        raise ValueError(preset_name)